import 'dart:async';

/**
 * Copyright 2019 - MagabeLab (Tanzania). All Rights Reserved.
 * Author Edwin Magabe    edyma50@yahoo.com
 */

import 'package:petitparser/petitparser.dart';
import 'package:meta/meta.dart';

import '../simple_smart_scraper.dart';

const String _otherPermittedAttributeValueChars =
    '-:;.,/?&=%#_@\\\u00B7\u0300-\u036F\u203F-\u2040\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF'
    '\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001'
    '\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD';

class StripException implements Exception {
  final String message;
  final String output;
  StripException(this.message, this.output);

  @override
  String toString() {
    return 'StripException: {message: $message ,\n output: $output} ';
  }
}

class _Parsers with ParserMixin {}

final _parsers = _Parsers();

typedef Future<String> Finalizer(String output);

/// Parse any children elements except text and provided exceptionalTags
class _ChildrenElementsParser extends Parser {
  Set<String> _tags;
  Parser _mainParser;
  Set<String> exceptionalTags;
  @override
  _ChildrenElementsParser(this.exceptionalTags);

  Parser copy() {
    return _mainParser ?? _ChildrenElementsParser(this.exceptionalTags);
  }

  @override
  Result parseOn(Context context) {
    if (_tags == null) {
      _tags = _parsers.getElementTags(context.buffer);
    }

    if (_mainParser == null) {
      for (var tag in _tags) {
        var parser = _parsers
            .spaceOptional()
            .seq((_parsers.element(tag) |
                _parsers.elementStartTag(tag: tag, isClosed: true) |
                _parsers.elementStartTag(tag: tag)))
            .seq(_parsers.spaceOptional());

        if (_mainParser == null) {
          _mainParser = parser;
        } else {
          _mainParser = _mainParser.or(parser);
        }
      }
    }
    return (_mainParser != null)
        ? _mainParser.parseOn(context)
        : Failure<String>(
            context.buffer, context.buffer.length, 'No tag element found');
  }
}

/// Parse any elements (child or not) except text and provided exceptionalTags
class AnyElementParser extends Parser<String> {
  Set<String> exceptionalTags;
  @override
  AnyElementParser(this.exceptionalTags);

  Parser<String> copy() {
    return AnyElementParser(this.exceptionalTags);
  }

  @override
  Parser<String> flatten([String message]) {
    return _FlattenParser(this, message);
  }

  _mainParser(String input) {
    Set<String> tags = _parsers.getElementTags(input).where((tag) {
      return (exceptionalTags != null) ? !exceptionalTags.contains(tag) : true;
    }).toSet();

    var mainParser;
    for (var tag in tags) {
      var parser = _parsers
          .spaceOptional()
          .seq((_parsers.element(tag) |
              _parsers.elementStartTag(tag: tag, isClosed: true) |
              _parsers.elementStartTag(tag: tag)))
          .seq(_parsers.spaceOptional());

      if (mainParser == null) {
        mainParser = parser;
      } else {
        mainParser = mainParser.or(parser);
      }
    }
    return mainParser;
  }

  @override
  Result<String> parseOn(Context context) {
    final position = context.position;
    final buffer = context.buffer;
    if (position < buffer.length) {
      final input = buffer.substring(position);

      var value = _parsers.getParserResult(
        parser: _mainParser(buffer),
        input: input,
      );

      if (value.isNotEmpty) {
        final tag = _parsers.getTagFromElementStartTag(value);

        final isClosed =
            _parsers.elementStartTag(tag: tag, isClosed: true).accept(value);
        int newPosition;
        if (isClosed) {
          newPosition = buffer.indexOf(value, position) + value.length;
        } else {
          var startTag = _parsers.getParserResult(
              parser: _parsers.elementStartTag(tag: tag), input: value);
          if (startTag.isNotEmpty) {
            newPosition = buffer.indexOf(startTag, position) + startTag.length;
          } else {
            return Failure<String>(buffer, position, 'No element found.');
          }
        }
        return Success<String>(buffer, newPosition, value);
      } else {
        return Failure<String>(buffer, buffer.length, 'No element found.');
      }
    } else {
      return Failure<String>(
          buffer, buffer.length, 'No element found, position > buffer.length.');
    }
  }
}

/// Parse text except children elements
class ElementTextParser extends Parser<String> {
  String tag;
  ElementTextParser(this.tag);

  @override
  Parser<String> copy() {
    return ElementTextParser(tag);
  }

  @override
  Parser<String> flatten([String message]) {
    return _FlattenParser(this, message);
  }

  @override
  Result<String> parseOn(Context context) {
    final int position = context.position;
    final String buffer = context.buffer;

    if (position < buffer.length) {
      final String input = buffer.substring(position);
      var result = _parsers.getParserResult(
          parser: _parsers._qualifiedElement(tag), input: input);
      if (result.isEmpty) {
        return Failure<String>(
            buffer, position, 'No element with $tag tag found.');
      }
      final int index = buffer.indexOf(result, position);
      if (index == -1) {
        return Failure<String>(buffer, position,
            'Failed to parse, No element with $tag tag found.');
      }
      final newPosition = index + result.length;

      var value;
      try {
        value = _parsers.strip(result);
        final isText = (!_parsers.childrenElements().accept(value));
        if (isText) {
          return Success<String>(buffer, newPosition, value);
        } else {
          final children = _parsers.getParserResults(
              parser: _parsers.childrenElements(), input: value);
          final childrenCount = children.length;
          if (childrenCount == 1) {
            final isClosed = _parsers
                .elementStartTag(isClosed: true)
                .accept(children.elementAt(0));
            if (isClosed) {
              return Failure<String>(buffer, newPosition,
                  '$tag does not contain text, but a closed element \n>>> ${children.elementAt(0)} <<<');
            } else {
              return Failure<String>(buffer, newPosition,
                  '$tag does not contain text, but an element \n>>> ${children.elementAt(0)} <<<');
            }
          } else if (childrenCount > 1) {
            return Failure<String>(buffer, newPosition,
                '$tag does not contain text, but $childrenCount elements \n>>> $children <<<');
          } else {
            return Failure<String>(
                buffer, newPosition, '$tag does not contain any text');
          }
        }
      } on StripException catch (_) {
        final isClosed =
            _parsers.elementStartTag(isClosed: true).accept(_.output);
        if (isClosed) {
          return Failure<String>(buffer, newPosition,
              '$tag is a closed element, therefore does not contain text \n>>> ${_.output} <<<');
        } else {
          return Failure<String>(buffer, newPosition, '$tag striping failed.');
        }
      }
    } else {
      return Failure<String>(buffer, buffer.length,
          'Failed to get text, position > buffer.length.');
    }
  }
}

class _FlattenParser extends DelegateParser<String> {
  final String message;
  _FlattenParser(Parser<String> delegate, [this.message]) : super(delegate);

  @override
  Result<String> parseOn(Context context) {
    final result = delegate.parseOn(context);
    if (result is Success<String>) {
      return result;
    } else {
      if (message == null) {
        return result;
      } else {
        return result.failure(message);
      }
    }
  }

  @override
  int fastParseOn(String buffer, int position) {
    return delegate.fastParseOn(buffer, position);
  }

  @override
  bool hasEqualProperties(FlattenParser other) =>
      identical(this, other) ||
      other is _FlattenParser &&
          runtimeType == other.runtimeType &&
          message == other.message;

  @override
  _FlattenParser copy() => _FlattenParser(delegate);
}

///Any parser that wishes to parse any formatted string(xml ,html etc) must mix with this mixin
mixin ParserMixin {
  Parser start() => char('<');

  Parser end() => char('>');

  Parser slash() => char('/');

  Parser spaceOptional() => whitespace().star();

  Parser quote() => char('"') | char("'");

  Parser equal() => char('=');

  Parser repeat(Parser p, int times) => repeatRange(p, min: times, max: times);

  Parser repeatRange(
    Parser p, {
    @required int min,
    @required int max,
  }) =>
      spaceOptional().seq(p).seq(spaceOptional()).repeat(min, max);

  /// In the following examples
  ///    final str = '''
  ///          chura CHURA ChurA ChUrA mkia is awesome
  ///       ''';
  ///
  /// nonCaseSensitiveChars("chura").flatten().matchesSkipping(str);
  /// returns [chura CHURA ChurA ChUrA]
  Parser nonCaseSensitiveChars(String str) {
    if ((str == null) || str.isEmpty) {
      return undefined('argument "str" can not be empty or null');
    }
    Parser p;
    str.split('').forEach((c) {
      final parser = pattern('${c.toLowerCase()}${c.toUpperCase()}');
      if (p == null) {
        p = parser;
      } else {
        p = p.seq(parser);
      }
    });
    return p;
  }

  Parser attributeValue() => (letter() |
          digit() |
          whitespace() |
          pattern(_otherPermittedAttributeValueChars))
      .star();

  Parser attributeKey([String key]) => (key == null)
      ? letter().plus().seq((letter() | digit() | pattern('-_~.')).star())
      : nonCaseSensitiveChars(key);

  Parser attribute([String attr]) => (attributeKey(attr)
      .seq(spaceOptional())
      .seq(equal())
      .seq(spaceOptional())
      .seq(word().plus().or(quote().seq(attributeValue()).seq(quote()))));

  /// In the following examples
  ///    final str = '''
  ///          <tag attr1 ="attribute1"> Text </tag>
  ///          <TAG> TEXT </TAG>
  ///            <tag/>
  ///       ''';
  ///  elementStartTag(tag:'tag'); matches both  <tag attr1 ="attribute1"> and  <TAG>
  ///     while
  ///  elementStartTag(tag:'tag',isClosed: true); matches only <tag/>
  ///  specify maxNoOfAttributes < 0 or null  for no attributes
  Parser elementStartTag({
    String tag,
    int maxNoOfAttributes = 6,
    bool isClosed = false,
  }) {
    final Parser attr = (maxNoOfAttributes <= 0 || (maxNoOfAttributes == null))
        ? spaceOptional()
        : spaceOptional().seq(attribute().star()).repeat(1, maxNoOfAttributes);

    Parser p = start()
        .seq((tag == null) ? word().plus() : nonCaseSensitiveChars(tag))
        .seq(spaceOptional())
        .seq(attr)
        .seq(spaceOptional());

    if (isClosed) {
      p = p.seq(slash());
    }

    p = p.seq(end());

    return p;
  }

  Parser elementEndTag([String tag]) => start()
      .seq(slash())
      .seq((tag == null) ? word().plus() : nonCaseSensitiveChars(tag))
      .seq(end());

  /// In the following examples
  ///    final str = '''
  ///          <tag attr1 ="attribute1"> Text </tag>
  ///          <TAG> TEXT </TAG>
  ///       ''';
  ///  element('tag'); matches both  <tag attr1 ="attribute1"> Text </tag>
  ///  and  <TAG> TEXT </TAG>
  Parser element(
    String tag, {
    Parser startTag,
    Parser endTag,
  }) {
    return spaceOptional().seq(
        ((startTag != null) ? startTag : elementStartTag(tag: tag))
            .seq(spaceOptional())
            .seq(spaceOptional()
                .seq(any()
                    .starLazy((endTag != null) ? endTag : elementEndTag(tag))
                    .flatten('element: Expected any text'))
                .pick(1)
                .optional(''))
            .seq(spaceOptional())
            .seq((endTag != null) ? endTag : elementEndTag(tag)));
  }

  ///Parse element with no attributes and text
  Parser emptyElement(String tag) {
    final closed =
        elementStartTag(tag: tag, maxNoOfAttributes: 0, isClosed: true);
    final complete = elementStartTag(tag: tag, maxNoOfAttributes: 0)
        .seq(spaceOptional())
        .seq(elementEndTag(tag));

    final br = start()
        .seq(spaceOptional())
        .seq(nonCaseSensitiveChars('br'))
        .seq(spaceOptional())
        .seq(end());
    return spaceOptional().seq(closed.or(complete).or(br));
  }

  /// Match any children elements except text and provided exception
  Parser childrenElements({Set<String> except = const <String>{}}) =>
      _ChildrenElementsParser(except);

  /// Match  any element except text and provided exception
  Parser anyElement({Set<String> except = const <String>{}}) =>
      AnyElementParser(except);

  /// In the following examples
  ///    final str = '''
  ///        <tr>  <tag attr1 ="attribute1"> Text </tag> </tr>
  ///          <TAG> TEXT </TAG>
  ///       ''';
  ///  parentElement("tr",element('tag'));
  ///  matches  <tr>  <tag attr1 ="attribute1"> Text </tag> </tr>
  Parser parentElement(
    String tag,
    Parser element, {
    Parser startTag,
    Parser endTag,
  }) {
    return spaceOptional().seq(
        ((startTag != null) ? startTag : elementStartTag(tag: tag))
            .seq(spaceOptional())
            .seq(element)
            .seq(spaceOptional())
            .seq((endTag != null) ? endTag : elementEndTag(tag)));
  }

  Parser parentElementWithNoAttributes(String tag) {
    return spaceOptional()
        .seq(elementStartTag(tag: tag, maxNoOfAttributes: 0))
        .seq(spaceOptional())
        .seq(spaceOptional()
            .seq(any()
                .starLazy(elementEndTag(tag))
                .flatten('element: Expected any text'))
            .pick(1)
            .optional(''))
        .seq(spaceOptional())
        .seq(elementEndTag(tag));
  }

  ///Parse and return a text of an element
  Parser elementText(String tag) => ElementTextParser(tag);

  ///Prefer using this method instead of a corresponding parser
  String getElementText({
    @required String tag,
    @required String input,
  }) {
    if ((tag == null) || (tag.isEmpty) || (input == null) || (input.isEmpty)) {
      return '';
    }
    return getParserResult(parser: elementText(tag), input: input);
  }

  ///Return the first successful parser result as a string
  String getParserResult({
    @required Parser parser,
    @required String input,
  }) {
    if ((parser == null) || (input == null) || (input.isEmpty)) {
      return '';
    }

    var results = getParserResults(parser: parser, input: input);
    return results.isNotEmpty ? results.elementAt(0).trim() : '';
  }

  ///Return all successful parser results
  List<String> getParserResults({
    @required Parser parser,
    @required String input,
  }) {
    return (parser == null)
        ? <String>[]
        : parser.flatten().matchesSkipping(input);
  }

  /// Example : _qualifiedElement('a');  parses
  ///  <a href="link">  , <a href="link"></a> and  <a href="link"/> successfully
  Parser _qualifiedElement(String tag) => element(tag)
      .or(elementStartTag(tag: tag, isClosed: true))
      .or(elementStartTag(tag: tag));

  String getAttributeValue({
    @required String tag,
    @required String attribute,
    @required String input,
  }) {
    if ((tag == null) ||
        tag.isEmpty ||
        (attribute == null) ||
        attribute.isEmpty ||
        (input == null) ||
        input.isEmpty) {
      return '';
    }

    final String element =
        getParserResult(parser: _qualifiedElement(tag), input: input);

    final String startTag = getParserResult(
        parser: elementStartTag(tag: tag)
            .or(elementStartTag(tag: tag, isClosed: true)),
        input: element);

    String attributeStr =
        getParserResult(parser: this.attribute(attribute), input: startTag);

    if (attributeStr.isNotEmpty) {
      final int index = attributeStr.indexOf('=', 0);
      if (index != -1) {
        var attr = attributeStr.substring(index + 1).trim();
        if (attr.isNotEmpty) {
          return removeQuotes(attr);
        } else {
          return attr;
        }
      }
    }
    return '';
  }

  bool beginAndEndWithTest(int quoteCode, String input) {
    return ((quoteCode == input[0].codeUnitAt(0)) &&
        (quoteCode == input[input.length - 1].codeUnitAt(0)));
  }

  bool beginAndEndWithDoubleQuote(String input) {
    return beginAndEndWithTest('"'.codeUnitAt(0), input);
  }

  bool beginAndEndWithSingleQuote(String input) {
    return beginAndEndWithTest("'".codeUnitAt(0), input);
  }

  String removeQuotes(String input) {
    if (beginAndEndWithSingleQuote(input) ||
        beginAndEndWithDoubleQuote(input)) {
      //remove the last quote
      var lastRemoved = input.replaceRange(input.length - 1, input.length, '');
      return lastRemoved.substring(1).trim(); //the remove the first quote
    } else {
      return input.trim();
    }
  }

  Set<String> getElementTags(String input) {
    var tags = <String>{};

    var elementsStartTags = getParserResults(
      parser: elementStartTag().or(elementStartTag(isClosed: true)),
      input: input,
    );

    for (var elemStartTag in elementsStartTags) {
      final tag = getTagFromElementStartTag(elemStartTag);
      if (!tags.contains(tag)) {
        tags.add(tag);
      }
    }
    return tags;
  }

  Map<String, String> getElementAttributes({
    @required Parser parser,
    @required String input,
  }) {
    var attributes = getParserResults(
        parser: attribute(),
        input: getParserResult(
            //only use start tag of parent element to avoid child-parent attribute mixing in the final results
            parser: elementStartTag().or(elementStartTag(isClosed: true)),
            input: getParserResult(
              parser: parser,
              input: input,
            )));

    Map<String, String> map = <String, String>{};
    for (var attr in attributes) {
      attr = attr.trim();
      if (attr.isNotEmpty) {
        final index = attr.indexOf('=', 0);
        if (index >= 0) {
          final key = attr.substring(0, index).trim();
          final value = removeQuotes(attr.substring(index + 1));
          if (key.isNotEmpty && value.isNotEmpty) {
            map[key] = value;
          }
        }
      }
    }
    return map;
  }

  bool hasAttribute({
    @required String tag,
    @required String attribute,
    @required String input,
  }) {
    return getAttributeValue(
      tag: tag,
      attribute: attribute,
      input: input,
    ).isNotEmpty;
  }

  ///1. If both tag and attributes are NOT NULL and NOT EMPTY, then attributes specified will be removed from the tag
  ///2. If the tag is NULL or EMPTY and attributes is NOT NULL and NOT EMPTY, then attributes specified will be removed from all tags
  ///3. if  the tag is NOT NULL and NOT EMPTY and attribute is NULL or EMPTY , then if the tag specified has attributes, they will all be removed
  ///4. If both tag and attributes are NULL or EMPTY , then all attributes of all tags will be removed
  String removeAttributes({
    String tag,
    Set<String> attributes,
    @required String input,
  }) {
    var output = input;
    final tags =
        ((tag != null) && tag.isNotEmpty) ? [tag] : [...getElementTags(input)];

    for (var currentTag in tags) {
      final elementList =
          getParserResults(parser: _qualifiedElement(currentTag), input: input);

      for (var elem in elementList) {
        final previous = elem;
        final currentTagAttributes = getElementAttributes(
                parser: _qualifiedElement(currentTag), input: elem)
            .keys;

        final attributesToRemove =
            ((attributes != null) && attributes.isNotEmpty)
                ? attributes
                : {...currentTagAttributes};
        for (var attrToRemove in attributesToRemove) {
          final attrFoundList =
              getParserResults(parser: attribute(attrToRemove), input: elem);
          for (var attr in attrFoundList) {
            elem = elem.replaceAll(attr, '');
          }
        }
        output = output.replaceAll(previous, elem);
      }
    }
    return output;
  }

  /// The String form is  key="Value"
  Map<String, String> getElementAttributesStrings({
    @required String input,
  }) {
    var attributes = getParserResults(
        parser: attribute(),
        input: getParserResult(
            //only use start tag of parent element to avoid child-parent attribute mixing in the final results
            parser: elementStartTag().or(elementStartTag(isClosed: true)),
            input: input));

    Map<String, String> map = <String, String>{};
    for (var attr in attributes) {
      attr = attr.trim();
      if (attr.isNotEmpty) {
        final index = attr.indexOf('=', 0);
        if (index >= 0) {
          final key = attr.substring(0, index).trim();
          final value = removeQuotes(attr.substring(index + 1));
          if (key.isNotEmpty && value.isNotEmpty) {
            map[key] = attr;
          }
        }
      }
    }
    return map;
  }

  ///if the tag is NOT NULL and NOT EMPTY, then the operation is performed only on that tag ,otherwise the operation is performed on all tags of the input
  ///Specifying NULL or EMPTY attributes means remove all attributes
  ///The results tends to be better as filterLoop increase up to a reasonable number
  Future<String> keepAttributes({
    String tag,
    @required Set<String> attributes,
    @required String input,
    int filterLoop = 2,
    Finalizer finalizer,
  }) {
    if (attributes == null) {
      attributes = <String>{};
    }

    if (filterLoop > 6) {
      filterLoop = 6;
    }

    if (filterLoop < 0) {
      filterLoop = 2;
    }
    var output = input;
    final tags =
        ((tag != null) && tag.isNotEmpty) ? [tag] : [...getElementTags(input)];

    Completer<String> completer = Completer();

    runFilter(Function cb) {
      Future.delayed(Duration(milliseconds: 80), () {
        int loop = 0;
        while (loop < filterLoop) {
          cb();
          loop++;
        }

        List<Parser> emptyElements =
            tags.map((tag) => emptyElement(tag)).toList();

        //remove empty elements
        int removeLoop = 0;
        while (removeLoop < (filterLoop + 1)) {
          output = remove(parsers: emptyElements, input: output);
          removeLoop++;
        }

        if (finalizer != null) {
          completer.complete(finalizer(output));
        } else {
          completer.complete(output);
        }
      });
    }

    runFilter(() {
      for (var currentTag in tags) {
        var elementList = getParserResults(
            parser: _qualifiedElement(currentTag), input: output);

        int currentElement = 0;
        while (currentElement < elementList.length) {
          var elem = elementList.elementAt(currentElement);

          final int elementStartPositionInTheOutput = output.indexOf(elem);
          // final elementEndPositionInTheOutput = elementStartPositionInTheOutput + elem.length;

          final currentTagAttributes = getElementAttributes(
                  parser: _qualifiedElement(currentTag), input: elem)
              .keys;
          final attributesToRemove = currentTagAttributes.where(
              (attributeToKeep) => !attributes.contains(attributeToKeep));

          for (var attrToRemove in attributesToRemove) {
            final attrFoundList =
                getParserResults(parser: attribute(attrToRemove), input: elem);
            for (var attr in attrFoundList) {
              final int attrStartIndex =
                  output.indexOf(attr, elementStartPositionInTheOutput);
              if (attrStartIndex != -1) {
                final int attrEndIndex = attrStartIndex + attr.length;
                //remove attribute and update output
                output = output.replaceRange(attrStartIndex, attrEndIndex, '');
              }
            }
          }

          //reload list with the current state/output
          elementList = getParserResults(
              parser: _qualifiedElement(currentTag), input: output);

          currentElement++;
        }
      }
    });

    return completer.future;
  }

  String remove({
    @required List<Parser> parsers,
    @required String input,
  }) {
    if (((parsers == null) || parsers.isEmpty) ||
        ((input == null) || input.isEmpty)) {
      return input;
    }
    String output = input;

    for (var parser in parsers) {
      var lst = getParserResults(parser: parser, input: input);
      for (var result in lst) {
        output = output.replaceAll(result, '');
      }
    }
    return output.trim();
  }

  int getParserResultCount({
    @required Parser parser,
    @required String input,
  }) {
    return getParserResults(parser: parser, input: input).length;
  }

  int getElementOpenStartTagCount(String input) {
    return getParserResultCount(parser: elementStartTag(), input: input);
  }

  int getElementEndTagCount(String input) {
    return getParserResultCount(parser: elementEndTag(), input: input);
  }

  bool verifyInput(String input) {
    if (getElementOpenStartTagCount(input) == getElementEndTagCount(input)) {
      final startTags = getParserResults(
        parser: elementStartTag(),
        input: input,
      ).map((s) => getTagFromElementStartTag(s));

      final endTags = getParserResults(
        parser: elementEndTag(),
        input: input,
      ).map((s) => getTagFromElementEndTag(s));
      for (int i = 0; i < startTags.length; i++) {
        if (!endTags.contains(startTags.elementAt(i))) {
          return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  String getTagFromElementStartTag(String input) {
    return getParserResult(
            parser: start()
                .seq(spaceOptional())
                .seq(letter().seq((letter() | digit()).star()))
                .seq(spaceOptional()),
            input: input)
        .replaceAll('<', '')
        .trim();
  }

  String getTagFromElementEndTag(String input) {
    return getParserResult(
            parser: start()
                .seq(spaceOptional())
                .seq(slash())
                .seq(spaceOptional())
                .seq(letter().seq((letter() | digit()).star()))
                .seq(spaceOptional())
                .seq(end()),
            input: input)
        .replaceAll('<', '')
        .replaceAll('/', '')
        .replaceAll('>', '')
        .trim();
  }

  ///Remove the parentElement and returns its children/content
  String strip(String input) {
    if (input == null || input.isEmpty) {
      throw StripException('Unexpected error occured', '');
    }
    input = input.trim();

    if (!verifyInput(input)) {
      throw StripException(
          '>>>>>\n***Error: parser error , invalid input formart **** <<<<<',
          input);
    }

    var closed = elementStartTag(maxNoOfAttributes: 0, isClosed: true)
        .or(elementStartTag(isClosed: true));
    var closedElements = getParserResults(parser: closed, input: input);
    for (var closedElem in closedElements) {
      final int index = input.indexOf(closedElem);
      final bool startedWithClosedElement = (index == 0);
      //strip failed
      if (startedWithClosedElement) {
        throw StripException(
            '>>>>>\n***Error:No one root parent element found, the first element is the closed element **** <<<<<',
            input);
      }
    }

    final startTag =
        elementStartTag(maxNoOfAttributes: 0).or(elementStartTag());

    var startElements = getParserResults(parser: startTag, input: input);
    for (var startElem in startElements) {
      if (input.indexOf(startElem) == 0) {
        final tag = getTagFromElementStartTag(startElem);
        final endTags =
            getParserResults(parser: elementEndTag(tag), input: input);
        int tagLastIndex = -1;
        int tagLen = 0;
        String endTag = '';
        for (var end in endTags) {
          int i = input.indexOf(end, tagLastIndex + 1);
          if (i > tagLastIndex) {
            tagLastIndex = i;
            tagLen = end.length;
            endTag = end;
          }
        }

        final elements = getParserResults(parser: element(tag), input: input);
        String firstElement;
        for (var e in elements) {
          if (input.indexOf(e) == 0) {
            firstElement = e;
            break;
          }
        }

        if (firstElement != null) {
          final bool isFirstElementParent =
              (((tagLastIndex + tagLen) == input.length) &&
                  (firstElement.length == input.length));

          final bool isFirstElementSibling =
              (((tagLastIndex + tagLen) < input.length) &&
                  (firstElement.length == (tagLastIndex + tagLen)));

          if (isFirstElementParent) {
            var striped =
                input.replaceAll(startElem, '').replaceAll(endTag, '').trim();

            return striped;
          }

          if (isFirstElementSibling) {
            throw StripException(
                '>>>>>\n***Error: Can not strip, The element*** \n\n$firstElement  \n\n\t***is not parent but sibling with*** \n\n${input.substring(firstElement.length)}\n\n<<<<<',
                input);
          }
        } else {
          throw StripException(
              '>>>>>\n***Error: parser error , invalid input formart **** <<<<<',
              input);
        }
      } else {
        throw StripException(
            '>>>>>\n***Error: No one root parent element found, siblings element found **** <<<<<',
            input);
      }
    }

    throw StripException('Unexpected error occured', input);
  }

  String stripRepeat(String input, int times) {
    if (!verifyInput(input)) {
      throw StripException(
          '>>>>>\n***Error: parser error , invalid input formart **** <<<<<',
          input);
    }
    if (times < 0) {
      times = 1;
    }

    if (times > 12) {
      times = 12;
    }
    for (int i = 0; i < times; i++) {
      input = strip(input);
    }
    return input;
  }

  Parser meta() => elementStartTag(tag: 'meta')
      .or(elementStartTag(tag: 'meta', isClosed: true));

  bool isMetaTag(String input) => meta().accept(input);

  Parser link() => elementStartTag(tag: 'link')
      .or(elementStartTag(tag: 'link', isClosed: true));

  bool isLinkTag(String input) => link().accept(input);

  Parser script() => elementStartTag(tag: 'script')
      .or(elementStartTag(tag: 'script', isClosed: true))
      .or(element('script'));

  bool isScriptTag(String input) => script().accept(input);

  ///Uses the strip method behind the scenes, therefore it only remove the tags and attributes any texts will be left behind
  /// to remove the entire element , use remove
  Future<String> removeTags({
    @required Set<String> tags,
    @required String input,
    Finalizer finalizer,
  }) async {
    if (((tags == null) || tags.isEmpty) ||
        ((input == null) || input.isEmpty)) {
      return (input != null)
          ? ((finalizer != null) ? finalizer(input.trim()) : input.trim())
          : ((finalizer != null) ? finalizer('') : "");
    }

    Completer<String> completer = Completer();

    String output = input;
    var keepTags =
        [...getElementTags(input)].where((tag) => !tags.contains(tag)).toList();

    Future<String> finalize() async {
      output = await removeEmptyTags(output, keepTags: keepTags);
      if (finalizer != null) {
        return finalizer(output);
      } else {
        return output;
      }
    }

    Future.delayed(Duration(milliseconds: 80), () async {
      for (var currentTag in tags) {
        var elemList = getParserResults(
            parser: _qualifiedElement(currentTag), input: output);

        if (elemList.isEmpty) {
          continue;
        }

        var currentElemIndex = 0;
        update() {
          elemList = getParserResults(
              parser: _qualifiedElement(currentTag), input: output);
          currentElemIndex = 0;
        }

        while (currentElemIndex < elemList.length) {
          var currentElem = elemList.elementAt(currentElemIndex);
          final int start = output.indexOf(currentElem);
          final int end = start + currentElem.length;
          var striped;
          try {
            striped = strip(currentElem);
          } on StripException catch (_) {
            final bool isClosed =
                elementStartTag(isClosed: true).accept(currentElem);
            if (isClosed) {
              output = output.replaceRange(start, end, '');
              update();
            } else {
              if (isMetaTag(currentElem) || isLinkTag(currentElem)) {
                output = output.replaceRange(start, end, '');
                update();
              } else {
                currentElemIndex++;
              }
            }
            continue;
          }

          if (currentElem != striped) {
            output = output.replaceRange(start, end, striped);
          }
          update();
        }
      }

      output = await finalize();
      completer.complete(output);
    });

    return completer.future;
  }

  ///Uses the strip method behind the scenes, therefore it only remove the tags and attributes any texts will be left behind
  /// to remove the entire element , use remove
  Future<String> keepTags({
    @required Set<String> tags,
    @required String input,
    Finalizer finalizer,
  }) async {
    if (((input == null) || input.isEmpty)) {
      return (input != null)
          ? ((finalizer != null) ? finalizer(input.trim()) : input.trim())
          : ((finalizer != null) ? finalizer('') : "");
    }

    if (tags == null) {
      tags = <String>{};
    }

    Set<String> tagsToRemove =
        [...getElementTags(input)].where((tag) => !tags.contains(tag)).toSet();

    return removeTags(tags: tagsToRemove, input: input, finalizer: finalizer);
  }
}
